package de.fraunhofer.iem.dataprovider.tools.ort.service

import de.fraunhofer.iem.dataprovider.configuration.OpenCodeApiProperties
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.entity.ToolEntity
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.tools.ort.dto.VulnerabilityDto
import de.fraunhofer.iem.dataprovider.tools.ort.json.OrtJson
import de.fraunhofer.iem.dataprovider.tools.ort.json.ResultJson
import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.engine.cio.*
import io.ktor.client.plugins.contentnegotiation.*
import io.ktor.client.request.*
import io.ktor.client.statement.*
import io.ktor.serialization.kotlinx.json.*
import jakarta.annotation.PreDestroy
import kotlinx.serialization.json.Json
import org.springframework.stereotype.Service

@Service
class OrtService(private val openCodeApiProperties: OpenCodeApiProperties, private val toolService: ToolService) {

    private val logger = getLogger(javaClass)

    @PreDestroy
    private fun destroy() {
        logger.info("ORT service shutdown. Closing HTTP client.")
        client.close()
    }

    private val client = getHttpClient()

    private fun getHttpClient(): HttpClient {
        return HttpClient(CIO) {
            install(ContentNegotiation) {
                json(
                    Json { ignoreUnknownKeys = true }
                )
            }
        }
    }

    val toolEntity: ToolEntity = getOrCreateToolEntity()


    /**
     * Queries the ORT API and transforms the received JSON objects into
     * VulnerabilityDtos.
     *
     * When the API call fails, or contains malformed/incomplete
     * elements this function returns an empty list.
     */
    suspend fun getOrtResults(repoId: Long): List<VulnerabilityDto> {

        val ortResults = queryOrtApi(repoId)
        logger.info("Got ${ortResults.size} ORT results for $repoId.")
        return ortResults.mapNotNull {
            if (it.cveId != null && it.packageName != null && it.severity != null) {
                VulnerabilityDto(it.cveId, it.packageName, it.severity)
            } else {
                null
            }
        }
    }

    private final fun getOrCreateToolEntity(): ToolEntity {
        val createToolDto = CreateToolDto("ORT")
        return toolService.findOrCreateTool(createToolDto)
    }

    private suspend fun queryOrtApi(repoId: Long): List<ResultJson> {
        return try {
            logger.info("Query ORT API for repo with id $repoId")
            val response: HttpResponse = client.get(getToolApiPath(repoId))
            val ortJson = response.body<OrtJson>()
            if (ortJson.code != 200) {
                throw Exception("ORT API returned with code ${ortJson.code}")
            }
            ortJson.result
        } catch (exception: Exception) {
            logger.error("Query to ORT API failed with exception $exception")
            emptyList()
        }
    }

    private fun getToolApiPath(repoId: Long): String {
        return "${openCodeApiProperties.basePath}$repoId${openCodeApiProperties.ort}"
    }
}
